#include <iostream>
#include <algorithm>

#include "inst.h"
#include "sol.h"
#include "fgrasp.h"
#include "lsearch.h"

static const int MAX_LS_TRIES = 20;

template <class T, class Sol>
Sol* Grasp<T,Sol>::solve()
{
  init();
  while (!stop_condition()) {
    constructor(get_alpha());
    local_search();
    update();
  }
  terminate();
  return sol;
}

template <class T,class Sol>
void Grasp<T,Sol>::update(void)
{
  if (*sol < *best) {
    *best = *sol;
    if (verbose()) cout << *best;
  }
  iter++;
}

/** This function returns the number of the current iteration.
   Iteration numbers are incremented after function `update'
   is called. */
template <class T, class Sol>
int Grasp<T,Sol>::iter_num()
{
  return iter;
}
const int ALPHA = 20;
  
FGrasp::FGrasp(FAP_inst &i, int m) :
Grasp<FAP_inst,FAP_sol>(i), max_iter(m), bc(10) { }

FGrasp::FGrasp(FAP_inst &i, FAP_sol &s, int m) :
Grasp<FAP_inst,FAP_sol>(i, s), max_iter(m), bc(10) { }

FGrasp::~FGrasp() {}

// update list of blocked frequencies
void FGrasp::update_bc(int trans, int freq)
{
  // find the nodes which were affected: only the
  // adjacent nodes.
  int i, j;
  Vector<Ord_Bc>::iterator it;
  for (i=1; i<=inst.num_trans(); i++) {
    int dist = inst.min_dist(inst.cell(trans), inst.cell(i));
    if (i != trans && dist > 0)
      // find channels now blocked in i because of freq:
      for (j = freq-dist; j<=freq+dist; j++) 
	      bc[i].block(j);
	}
	// remove transmitter trans from the vector
	for (it=ord_bc.begin(); it!=ord_bc.end(); it++) {
	  if (((*it).p - &bc[0]) == trans) {
	    ord_bc.erase(it);
	    break;
	  }
	}
  // now, reorder the vector.
  sort(ord_bc.begin()+1, ord_bc.end());
}

int FGrasp::choose_channel(int trans)
{
  // find the first channel non blocked in trans
  int j, nf = inst.num_freq();
  for (j=1; j<=nf  && bc[trans].blocked(j); j++)
    ;
    
  if (j != inst.num_freq()+1) {
    return j;
  } else {
    // there is no available channel.
    return (rand() % inst.num_freq()) + 1;
  }  
}

int FGrasp::choose_trans(int alpha)
{
  // first, choose a position in rcl
  int remaining_trans = ord_bc.size();
  float perc = alpha/100.0;
  int rcl_size = (int)(perc * remaining_trans);
  if (rcl_size == 0) rcl_size++;
  int r = (rand()% (rcl_size))+1;

  // now return it
  int pos = ord_bc[r].p - bc.begin();
  return pos;
}

void FGrasp::init_constructor()
{
  bc.clear();
  bc.reserve(inst.num_trans()+1);
  ord_bc.clear();
  int i;
  for (i=0; i<=inst.num_trans(); i++) {
    bc.push_back(Bc());
    ord_bc.push_back(Ord_Bc(bc[i]));
  }
  sort(ord_bc.begin()+1, ord_bc.end());
}

void FGrasp::constructor(int alpha)
{
  init_constructor();
  // the first transmitter is random
  int n = inst.num_trans();
  int i, trans = (rand() % n) + 1;

  for (i=0;;) {
    int freq = choose_channel(trans);
    sol->set_channel(trans, freq);
    update_bc(trans, freq);
    if (++i >= n) break;
    trans = choose_trans(alpha);
  }
  sol->calc_of();
  if (verbose()) (cout << " gc ").flush();
}

void FGrasp::local_search()
{
  FLocal_search ls(inst, sol, bc, MAX_LS_TRIES);
  ls.solve();
  if (verbose()) (cout << " ls ").flush();
}

void FGrasp::update()
{
  Grasp<FAP_inst,FAP_sol>::update();
  if (verbose())  cout << " iter[" << iter_num() << "]";
}

