#include <iostream>
#include <string>
#include <fstream>

#include "inst.h"

/// dummy constructor
FAP_inst::FAP_inst() :
  n_freqs(80), n_trans(0), n_bs(0),
  d(0), co(0), ad(0), spectrum_ini(1)
{
}

/// This function reads a new instance of the FAP from file `ifile'.
/// Parameter iformat is true only if the instance is in the fap.zib.de
/// format.
/// @returns a pointer to the new instance
FAP_inst *
FAP_inst::create_instance(ifstream &ifile, bool iformat=false)
{
  FAP_inst *fi = new FAP_inst();
  if (!fi) return 0;
  if ( !fi->read_inst(ifile, iformat) ) {
    delete fi;
    return 0;
  }
  else
    return fi;
}


bool FAP_inst::read_inst(ifstream &ifile, bool iformat=false)
{
  if (iformat) {
    return read_iformat(ifile);
  } else {
    return read_rand_instance(ifile);
  }
  if (!symmetric())
    cout<<"Warnning: the instance is not symmetric!\n";
}

void FAP_inst::set_dist(int i, int j, int dist)
{
	b(i,j); exchange(i,j);
#ifdef VEC
	d[i].insert(mapi::value_type(j,dist));
#else
  d[i][j] = dist;
#endif  	
}
void FAP_inst::set_co_int(int i, int j, float interf)
{
	b(i,j); exchange(i,j);
#ifdef VEC
	co[i].insert(mapf::value_type(j,interf));
#else
  co[i][j] = interf;
#endif  	
}
void FAP_inst::set_ad_int(int i, int j, float interf)
{
	b(i,j); exchange(i,j);
#ifdef VEC
	ad[i].insert(mapf::value_type(j,interf));
#else
  ad[i][j] = interf;
#endif  	
}

int FAP_inst::min_dist(int i, int j)
{
  b(i,j); exchange(i,j);
#ifdef VEC
  mapi::iterator it = d[i].find(j);
	if (it == d[i].end())
	  return 0;
	else
	  return (*it).second;
#else
  return d[i][j];
#endif  	
}

float FAP_inst::co_int(int i, int j)
{
  b(i,j); exchange(i,j);
#ifdef VEC
  mapf::iterator it = co[i].find(j);
	if (it == co[i].end())
	  return 0;
	else
	  return (*it).second;
#else
  return co[i][j];
#endif  	
}

float FAP_inst::ad_int(int i, int j)
{
  b(i,j); exchange(i,j);
#ifdef VEC
  mapf::iterator it = ad[i].find(j);
	if (it == ad[i].end())
	  return 0;
	else
	  return (*it).second;
#else
  return ad[i][j];
#endif
}
void FAP_inst::init_vectors(int nb, int nt)
{
  int i;
  n_trans = nt;
  n_bs = nb;
  demand.clear(); trans_bs.clear();
  trans_bs.reserve (nt+1);
  demand.reserve(nb+1);
  for (i=0; i<=nb; i++)
    demand.push_back(0);
  for(i=0; i<=nt; i++)
    trans_bs.push_back(0);

#ifdef VEC
  co.clear(); ad.clear(); d.clear();
  co.reserve(nb+1);
  ad.reserve(nb+1);
  d.reserve(nb+1);
  for (int i=0; i<=nb; i++) {
    co.push_back(mapf());
    ad.push_back(mapf());	
    d.push_back(mapi());
  }
#else
  co = new float*[nb+1];
  ad = new float*[nb+1];
  d = new int*[nb+1];
  for (int i=0; i<=nb; i++) {
    co[i] = new float[nb+1];
    ad[i] = new float[nb+1];
    d[i] = new int[nb+1];
  }
#endif
}

FAP_inst::~FAP_inst()
{
#ifndef VEC
  if (co && ad && d) {
    for (int i=0; i<=n_bs; i++) {
      delete [] co[i];
      delete [] ad[i];
      delete [] d[i];
    }
    delete [] co;
    delete [] ad;
    delete [] d;
  }
#endif
}

bool FAP_inst::read_rand_instance(ifstream &ifile)
{
  // read number of base stations 
  ifile >> n_bs;
  if (n_bs < 2) {
    return false;
  }

  // read number of transmitters
  ifile >> n_trans;
  if (n_trans < n_bs) {
    return false;
  }
  
  init_vectors(n_bs, n_trans);

  //  read demand
  int i, j, demand;
  for (i=j=1; i<=n_bs; i++) {
    ifile >> demand;
    if (demand < 1){
      cerr << "Error: demand must be at least 1\n";
      return false;
    }
    if (j+demand-1 > n_trans) {
      cerr << "Error: total demand is greater than stated in line 1\n";
      return false;
    }
    set_demand(i, demand);
    
    int sum = j+demand;
    for (; j<sum; j++) {
      set_trans_bs(j, i);
    }
  }
  // read distance matrix
  string s, error_msg("Error: wrong format for instance file");
  ifile >> s;
  if (s != "[DistanceMatrix]") {
    cerr << error_msg;
    return false;
  }

  int num;
  for (i=1; i<=n_bs; i++) {
    for (j=1; j<=n_bs; j++){
      ifile >> num;
      if (num < 0) {
	cerr << error_msg;
	return false;
      }
      set_dist(i, j, num);
    }
  }

  // read cochannel channel interf.
  ifile >> s;
  if (s != "[co-channelInterference]") {
    cerr << error_msg;
    return false;
  }

  float f;
  for (i=1; i<=n_bs; i++) {
    for (j=1; j<=n_bs; j++){
      ifile >> f;
      if (f < 0 || f > 1) {
	cerr << error_msg;
	return false;
      }
      // set interference between base stations i and j
      set_co_int(i, j, f);
    }
  }
  // read adjacent channel interf.
  ifile >> s;
  if (s != "[adj-channelInterference]") {
    cerr << error_msg;
    return false;
  }

  for (i=1; i<=n_bs; i++) {
    for (j=1; j<=n_bs; j++){
      ifile >> f;
      if (f < 0 || f > 1) {
	cerr << error_msg;
	return false;
      }
      set_ad_int(i, j, f);
    }
  }
  // read last part of file.
  ifile >> s;
  if (s != "[blockedFrequencies]") {
    cerr << error_msg;
    return false;
  }
  ifile >> s;
  if (s != "end") {
    cerr << error_msg;
    return false;
  }
  return true;
}

void FAP_inst::print()
{
  cout << "n freqs: " << n_freqs << " n trans: " << n_trans
       << " n bs: " << n_bs << "\ndemand:\n";
  int i, j;
  for (i=1; i<=n_bs; i++)
    cout << demand[i] << " ";
  cout << "\ndistances:\n";
  for (i=1; i<=n_bs; i++) {
    for (j=1; j<=n_bs; j++) {
      cout << min_dist(i,j) <<" ";
    }
    cout <<endl;
  }
  cout << "==== size of a is " << n_trans << endl;
}

/// tests if this instance is symmetric: the value of the
/// property p(x,y) is the same is p(y,x)
bool FAP_inst::symmetric()
{
  int i,j, n = num_bs();
  for (i=1; i<=n; i++)
    for (j=1; j<=n; j++)
      if (min_dist(i,j) != min_dist(j,i) ||
	  co_int(i,j) != co_int(j,i) ||
	  ad_int(i,j) != ad_int(j,i))
	return false;
  return true;
}
