#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <string>
#include <iostream>
#include <fstream>
#include <map>
#include <sys/timeb.h>

#include "inst.h"
#include "sol.h"
#include "fprel.h"
#include "common.h"

using namespace CO_common;

const int DEFAULT_MAX_TIME = 36000; // ten hours
const int HC_MAX_ITER = 200;
const int ALPHA = 2;

class FAP_app : public app {
  int num_run;     // number of runs the program will do
  int seed;        // ramdomizer seed;
  int max_time;    // maximum time the program should run.
  int nfreq;       // numb. of frequencies used.
  int look4;       // value sought
  bool opt_hc;     // hill climb is used?
  int max_iter; // number of hc iterations
  bool opt_tcol;   // is tcoloring used?
  bool opt_grasp;  //is grasp used?
  int alpha;       // alpha used (without reac. grasp)
  bool opt_reac_grasp;  // is react. grasp used?
  bool opt_path_relink; // is path relinking used?
  bool opt_reverse_pr;  // is reverse path relinking used?
  bool opt_final_pr;    // is final path relinking used?

  bool verbose;         // in verbose mode?
  bool hide_sol;        // don't show final solution
  bool show_stat;       // show statistics.
  bool show_histogram;  // show frequency histogram
  bool opt_help;        // show help message
  bool input_format;    // input instance in `scenario format'
  time_t init_time;
  ifstream *ifile;
  Vector<FAP_sol> results;  // generated solutions
  string file_name;
  string read_sol;
  FAP_sol *s;
  FAP_inst *fi;

  bool validate_cmdl();
  void print_help();
  void find_a_solution(FAP_inst &fi, FAP_sol &s);
public:
  // constructor
  FAP_app(int argc, char *argv[]) :
    app(argc, argv), num_run(1),
    seed(0), max_time(DEFAULT_MAX_TIME), nfreq(80), look4(-1),
    opt_hc(false), max_iter(HC_MAX_ITER), opt_tcol(false),
    opt_grasp(false), alpha(ALPHA), opt_reac_grasp(false),
    opt_path_relink(false), opt_reverse_pr(false),
    verbose(false), hide_sol(true), show_stat(false),
    show_histogram(false), opt_help(false), 
    input_format(false), init_time(time(0)), ifile(0), 
    read_sol(""), s(0), fi(0) {}

  // destructor
  virtual ~FAP_app();
  void add_cmdl_opts();
  bool process_cmdline();
  string prog_name() { return "FAP"; }
  ifstream *inst_file() { return ifile; }
  void solve_fap();
  inline bool time_is_over();
  void do_read_sol();
  FAP_sol *get_solution();
  // return true if the application should print messages to screen
  bool get_verbose() { return verbose; }
};

/* destructor: free memory allocated by program */
FAP_app::~FAP_app() 
{ 
  if (ifile) 
    delete ifile; 
  if (s) 
    delete s;
  if (fi)
    delete fi;
}

/** this structure has all information needed to parse the
   comand line. To add an option, just enter a new element. */
void FAP_app::add_cmdl_opts()
{
  // opts related to solution type
  parser.add_opt("DownHill heuristic", "-h", &opt_hc);
  parser.add_opt("run T-Coloring heuristic", "-tc", &opt_tcol);
  parser.add_opt("run GRASP", "-g", &opt_grasp);
  parser.add_opt("number of iterations", "-ni", &max_iter);
  parser.add_opt("value of alpha (not used with reactive GRASP)",
		 "-alpha", &alpha);
  parser.add_opt("value to look for", "-l4", &look4);
  parser.add_opt("run reactive GRASP", "-rg", &opt_reac_grasp);
  parser.add_opt("run path relinking (+ React.GRASP)",
		 "-pr", &opt_path_relink);
  parser.add_opt("run reverse Path Relinking (+ PR & React.GRASP)",
		 "-rpr", &opt_reverse_pr);
  parser.add_opt("final Path Relinking (+ PR & Rev.PR & React.GRASP)",
		 "-fpr", &opt_final_pr);

  // generic options
  parser.add_opt("run the algorithm <num> times", "-n", &num_run);
  parser.add_opt("max running time", "-time", &max_time);
  parser.add_opt("number of frequencies used", "-nfreqs", &nfreq);
  parser.add_opt("random number generator seed", "-seed", &seed);
  
  // miscellanius options
  parser.add_opt("read COST259 input format file", "-if", &input_format);
  parser.add_opt("enable verbose mode", "-v", &verbose);
  parser.add_opt("don't print final solution", "-np", &hide_sol);
  parser.add_opt("show statistics", "-stat", &show_stat);
  parser.add_opt("show frequency histogram",
		 "-hist", &show_histogram);
  parser.add_opt("show this help message",
		 "-help", &opt_help);
  parser.add_opt("read a solution and print the objective value",
			"-rs", &read_sol);
};

void FAP_app::print_help()
{
  // print usage and exit.
  parser.print_options();
  exit(0); // it's correct to exit here
}  

/*  function called when argument "-rs" is given.
   read an assignment from the file specified and
   print its value.*/
void FAP_app::do_read_sol()
{
  // read the instance
  FAP_inst *fi =
    FAP_inst::create_instance(*inst_file(), input_format);
  if (fi) {
    // create a new solution
    FAP_sol sol(*fi);
    // first open the file in which the solution is stored
    ifstream *infile = new ifstream(read_sol.c_str());
    if (!infile) {
      cerr << "error: cannot open solution file " << read_sol
		 << endl;
      exit(1);
    }
    // read the solution stored in infile
    sol.input_sol(*infile);
    // calculate the objective function for this instance
    sol.calc_of();
    // print the objective  value of this solution
    cout << sol;
    delete infile;
    // finish the program
    exit(0);
  } else {
    cerr << "error: problems creating instance\n";
    exit(1);
  }
}

bool FAP_app::process_cmdline()
{
  // if there was no argument
  if (parser.num_args() == 1) {
    print_help();
  }
  // the file name is always the first argument
  file_name = parser.get_arg(1);
  if (file_name[0] != '-'){ // if the first arg is not an option
    // tru to open input file
    ifile = new ifstream(file_name.c_str(), ios::in);
    if (!ifile) {
      cerr << "Error: file " << file_name
					 << " could not be opened.\n";
      exit(1);
    }
    parser.remove_arg(1);
  } else {  // is an isolated option 
    print_help();
    exit(0);
  }
  app::process_cmdline();
  // if option `-help' was present
  if (opt_help) {
    print_help();
    exit(0);
  }
  // if we have to just read a solution
  if (read_sol != "") 
	do_read_sol();
  return true;
}

/// Try to validade values passed by command line.
bool FAP_app::validate_cmdl()
{
  // number of frequencies
  if (nfreq < 1) {
    cerr << "Error: num freqs must be positiv.\n";
    return false;
  }
  if (opt_reac_grasp || opt_path_relink
      || opt_reverse_pr || opt_final_pr)
  {
    opt_grasp = true;
    if (opt_path_relink || opt_reverse_pr || opt_final_pr) {
      opt_reac_grasp = true;
      if (opt_reverse_pr || opt_final_pr) {
				opt_path_relink = true;
				if (opt_final_pr) opt_reverse_pr = true;
      }
    }
  } else if (!opt_tcol && !opt_hc) {
    opt_grasp = true;
  }
  return true;
}

/// returns true if the maximum time set to the application
/// is over.
bool FAP_app::time_is_over()
{
  if (difftime(time(0), init_time) > max_time)
    return true;
  else
    return false;
}

void FAP_app::find_a_solution(FAP_inst &fi, FAP_sol &s)
{
  FGrasp  *g = 0;
  if (opt_hc || opt_tcol) {
    cout << "Option still not implemented\n";
  } else if (opt_final_pr) {
    cout << "I'll start final path relinking\n";
    g = new FFinalPathRel(fi, s, max_iter);
  } else if (opt_reverse_pr) {
    cout << "I'll start reverse path relinking\n";
    g = new FReversePathRel(fi, s, max_iter);
  } else if (opt_path_relink) {
    cout << "I'll start path relinking\n";
    g = new FPathRel(fi, s, max_iter);
  } else if (opt_reac_grasp) {
    cout << "I'll start reac grasp\n";
    g = new FReactGrasp(fi, s, max_iter);
  } else if (opt_grasp) {
    g = new FGrasp(fi, s, max_iter);
  }
  if (!g) {
    cerr << "error: could not create a solver.\n";
    exit(1);
  }
  g->set_look4(look4);
  g->solve();
  delete g;
}
void FAP_app::solve_fap()
{
  // protect against fool programmers :-)
  if (!fi) {
    (cout << "reading instance ... ").flush();
    fi =  FAP_inst::create_instance(*inst_file(), input_format);
    (cout << "done.\n").flush();
  }
  if (!fi) {
    cout << "Error: problems creating instance\n";
     return;
  }
  // init the randomizer
  srand(time(0));
  s = new FAP_sol(*fi);
  if (!s) {
    cout << "Error: can not create a solution\n";
    return;
  }

  // put some information in a `.out' file
  string out_file_name = file_name+".out";
  ofstream *out_file
    = new ofstream(out_file_name.c_str(), ios::app);
  *out_file << "====================\n"
            << "started at ";

  // create a tmp file with the pid of this program
  string spid;
  int pid = getpid();
  while (pid > 0) {
		spid += (char)('0' + pid % 10);
		pid /= 10;
  }
  string tmp_file_name = string("/tmp/")+  spid + ".gfap";
  system((string("date >")+tmp_file_name).c_str());
  ifstream *tmp_file = new ifstream(tmp_file_name.c_str());
  while (!tmp_file->eof()) {
    string tmp_str;
    *tmp_file >> tmp_str;
    *out_file << tmp_str << ' ';
  }
  *out_file << endl;
  delete tmp_file;
  delete out_file;

  // run GRASP the required number of times
  for (int i=0; i<num_run; i++) {
    find_a_solution(*fi, *s);
    // write the solution to a file
    ofstream *os
      = new ofstream(out_file_name.c_str(), ios::app);
    if (!os) {
      cerr << "error openning output file. goodbye.\n";
      exit(1);
    }
    *os << *s;
    delete os;
    // append solution to a list:
    results.push_back(*s);
    init_time = time(0);
  }
  out_file
    = new ofstream(out_file_name.c_str(), ios::app);
  *out_file << "finished at ";
  system((string("date >")+tmp_file_name).c_str());
  tmp_file = new ifstream(tmp_file_name.c_str());
  while (!tmp_file->eof()) {
	string tmp_str;
	*tmp_file >> tmp_str;
	*out_file << tmp_str << ' ';
  }
  *out_file << endl;
  delete tmp_file;
  delete out_file;
}
  
static bool end_fap = false;
bool global_end_fap()
{
  return end_fap;
}

/* process a SIGINT (control-c) */
void signal_handler(int signum)
{
  if (signum == SIGINT){
    // in the first time, just set the flag
    if (end_fap == false) {
      cerr << "Sigint received!\n";
      end_fap = true;
    } else {
      // if end_fap is true, finish the application
      cerr << "Quitting the application!\n";
      exit(1);
    }
  }
}

FAP_app *fapp;

// global function
// return true if the application should print messages to screen
bool verbose(){
  return fapp->get_verbose();
}

bool time_is_over()
{
  return fapp->time_is_over();
}

int main(int argc, char *argv[])
{
  struct timeb t1,t2;
  cout << "GFAP: A GRASP to FAP, (c) 2002 - Carlos A.S. Oliveira\n";
  fapp = new FAP_app(argc, argv);
  if (!fapp) return 1;
  fapp->add_cmdl_opts();
  fapp->process_cmdline();
  // install signal handler for control-C
  signal(SIGINT, signal_handler);
  ftime(&t1);
  fapp->solve_fap();
  ftime(&t2);
  time_t t = t2.time - t1.time;
  unsigned short m;
  if (t2.millitm < t1.millitm) {
	  t--;
	  m =  100 + t2.millitm - t1.millitm;
  } else {
	  m =  t2.millitm - t1.millitm;
  }
  cout<< "total time: " << t << "." << m << "s\n";
  delete fapp;
  return 0;
}
