#include "rgrasp.h"

const int FReactGrasp::ALPHA_VEC_SIZE = 10;
const int FReactGrasp::UPDATE_ITERATIONS = 20;

void FReactGrasp::init_structures()
{
  // probability is a vector whose sum is equal to 1.
  // (is the probability distribution vector).
  probability.push_back(0);
  avg_val.push_back(0);
  for (int i=1; i<=ALPHA_VEC_SIZE; i++) {
    probability.push_back(1.0/ALPHA_VEC_SIZE);
    avg_val.push_back(sol->get_of());
  }
}

FReactGrasp::FReactGrasp(FAP_inst &i, int max_iter) :
  FGrasp(i, max_iter), alpha_freq(ALPHA_VEC_SIZE+1, 0),
  q(ALPHA_VEC_SIZE+1, 0), current_alpha(0)
{
  init_structures();
}

FReactGrasp::FReactGrasp(FAP_inst &i, FAP_sol &s, int max_iter) :
  FGrasp(i, s, max_iter), alpha_freq(ALPHA_VEC_SIZE+1, 0),
  q(ALPHA_VEC_SIZE+1, 0), current_alpha(0)
{
  init_structures();
}

FReactGrasp::~FReactGrasp() {}

/**
 This function returns a new value for alpha, based on the relative
 probabilities in the fg->prob vector.
 
 The choice of alpha is based in a set of probabilities p_i,
 i=1..n.  These probabilities are updated in
 `RGRASP-UPDATE-ITERATIONS' iterations.
 
 The new probabilities are given by:
 
      p[i] = q[i] / sum(from j=1 to m of q[j]),
      
 where

      q[j] = f(S')/A[j].
      
  @returns a number between 1 and 100 (the percentage of RCL to  be
  considered).  */
int FReactGrasp::get_alpha()
{
  int i; double sum;
  // prob is a probability between 0 and 1.
  double prob = (1+(rand()%1000))/1000.0;
  for (sum=0, i=1; sum<prob; i++)
    sum += probability[i];
  i--;
  current_alpha = i;
  return int(probability[i] * 100);
}

/** @name update of averages.
  This function updates the average value of the objective function
   found with this value of alpha. It's function after every
   call to grasp constructor. */
void FReactGrasp::update_avg()
{
  // position of the current alpha value in the alpha vector
  double old_value = avg_val[current_alpha];
  double new_value = sol->get_of();
  int n = alpha_freq[current_alpha];
  avg_val[current_alpha] = (old_value*n + new_value)/(n+1);
  alpha_freq[current_alpha]++;
}

/** update the value of alpha. This is done every
    RGRASP-UPDATE-ITERATIONS iterations. */
void FReactGrasp::update_alpha()
{
  int i; double sum = 0;
  for (i=1; i<=ALPHA_VEC_SIZE; i++){
    //old: // q[i] = pow(get_of()/avg_val[i], GRASP_DELTA);
    // the `power of' rule should be descard, according to Mauricio.
    q[i] = sol->get_of()/avg_val[i];
    sum += q[i];
  }
  // normalize the probability values.
  for (i=1; i<=ALPHA_VEC_SIZE; i++)
    probability[i] = q[i]/sum;
}
/** called at the end of each iteration if GRASP constructor.*/
void FReactGrasp::update()
{
  FGrasp::update();
  if ((iter_num() % UPDATE_ITERATIONS) == 0)
    update_alpha();
}
/** This constructor is modified to update the reactive GRASP data structures. */
void FReactGrasp::constructor(int alpha)
{
  FGrasp::constructor(alpha);
  update_avg();
}
