#include <iostream>
#include <map>
#include <math.h>

#include "inst.h"
#include "sol.h"
#include "iformat.h"

/* obs.: This code assumes that interferece is symetric: the
 *  interference from a to b is the same as from b to a.
 *  this makes difference in:
 *     inteference()
 *     interference_due(int trans)
 */

typedef multimap<int, int, less<int> > Freq_map;


FAP_sol::FAP_sol(FAP_inst &f) :
  fi(&f),  unf(0), interf(0),
  freq(f.num_trans()+1, 1)
{
  rand_sol();
  calc_of();
}

FAP_sol::FAP_sol(const FAP_sol &s) :
  fi(s.fi), unf(s.unf), interf(s.interf), freq(fi->num_trans()+1)
{
  int i;
  for (i=0; i<=fi->num_trans(); i++)
    freq[i] = s.freq[i];
  calc_of();
}

FAP_sol &FAP_sol::operator= (const FAP_sol &s)
{
  fi = s.fi;
  unf = s.unf; interf = s.interf;
  freq.clear();
  for (int i=0; i<=fi->num_trans(); i++) {
    int f = s.freq[i];
    freq.push_back(f);
  }
  return *this;
}

double FAP_sol::unfeasibility()
{
  int i, j, n = fi->num_trans();
  unf = 0;
  for (i=1; i<n; i++)
    for (j=i+1; j<=n; j++)
      if (abs(freq[i] - freq[j]) <
          fi->min_dist(fi->cell(i),fi->cell(j))) 
	      unf +=1;
  return unf;
}

double FAP_sol::interference()
{
  int i;
  Freq_map m;
  for (i=1; i<=fi->num_trans(); i++) {
    assert(freq[i] > 0 && freq[i] <= fi->num_freq());
    m.insert(Freq_map::value_type(freq[i], i));
  }
  
  // cochannel interference
  double csum = 0;
  pair<Freq_map::iterator, Freq_map::iterator> pi, pi2;
  Freq_map::iterator it, it2;
  for (i=1; i<=fi->num_freq(); i++)
  {
    // find the range of transmitters using this channel
    pi = m.equal_range(i);
    // test if its not empty
    if (pi.first != m.end())
      // iterate throug the combinations
      for (it = pi.first; it != pi.second;  it++)
	for (it2 = pi.first; it2 != pi.second; it2++)
	  // sum interference in just one direction
	  if (it->second < it2->second)
	    csum += fi->co_int(fi->cell(it->second),
			       fi->cell(it2->second));
  }
  
  // adj. channel interference
  double asum = 0;
  for (i=1; i<fi->num_freq(); i++)
  {
    // find the range of transmitters using channel i
    pi  = m.equal_range(i);
    // test if it is not empty
    if (pi.first != m.end())
    {
      // find the range of transmitters using channel i+1
      pi2 = m.equal_range(i+1);
      // test if it is not empty
      if (pi2.first != m.end())
	// sum the interference for each combination
	// of these transmitters
        for (it = pi.first; it != pi.second; it++)
	  for (it2 = pi2.first; it2 != pi2.second; it2++)
	    asum += fi->ad_int(fi->cell(it->second),
			       fi->cell(it2->second));
    }
  }
  interf = csum + asum;
  return interf;
}

void FAP_sol::rand_sol()
{
  assert(fi != 0);
  int i, n = fi->num_trans();
  for ( i=1; i <= n; i++)
    freq[i] = (::rand() % fi->num_freq()) + 1;
  calc_of();
}

#define of_less(u1, i1, u2, i2) \
  u1 < u2 ? true : ((u1 == u2 && i1 < i2) ? true : false)

/// less than operator
bool FAP_sol::operator< (const FAP_sol &s) const
{
  assert(fi != 0);
  return of_less(unf, interf, s.unf, s.interf);
}

/// Equal operator
bool FAP_sol::operator==(const FAP_sol &s) const
{
  if (fi != s.fi || unf != s.unf ||
      interf != s.interf || freq != s.freq)
    return false;
  return true;
}

bool verbose();

ostream &operator<<(ostream &o, FAP_sol &s)
{
   assert(s.fi != 0);
   // o.precision(2);  // doesn't work ...
   if (verbose()) {
     o << "(" << s.get_of() << ") " ;
   }
   o << "interf: ";
   o.width(8);
   o <<  s.interf << " unf: ";
   o.width(6);
   o << s.unf << endl;
   return o;
}

/// calculates and return the interference due only to one transmitter
double FAP_sol::interference_due(int trans)
{
  int i;
  Freq_map m;
  for (i=1; i<=fi->num_trans(); i++) {
    assert(freq[i] > 0 && freq[i] <= fi->num_freq());
    m.insert(Freq_map::value_type(freq[i], i));
  }

  // ----------------------
  // cochannel interference
  double csum = 0;
  pair<Freq_map::iterator, Freq_map::iterator> pi;
  Freq_map::iterator it;

  // we check just for the frequency `freq[trans]'
  pi = m.equal_range(freq[trans]);
  // test if the range is not empty
  if (pi.first != m.end())
    // compare with all others with same frequency
    for (it = pi.first; it != pi.second;  it++)
      // ensure the transmitters are different
      if (it->second != trans)
	csum += fi->co_int(fi->cell(it->second),
			   fi->cell(trans));

  // -------------------------
  // adj. channel interference
  double asum = 0;

  // find the range of  `freq[trans]+1'
  pi = m.equal_range(freq[trans]+1);
  // test it is not empty
  if (pi.first != m.end())
    // sum for each one in the range
    for (it = pi.first; it != pi.second; it++)
      asum += fi->ad_int(fi->cell(it->second),
			 fi->cell(trans));

  // same thing for `freq[trans]-1'
  pi = m.equal_range(freq[trans]-1);
  // test it is not empty
  if (pi.first != m.end())
    // sum for each one in the range
    for (it = pi.first; it != pi.second; it++)
      asum += fi->ad_int(fi->cell(it->second),
			 fi->cell(trans));

  return csum + asum;
}

/// calculates and return the unfesibility due only to one transmitter
double FAP_sol::unfeasibility_due(int trans)
{
  int j, n = fi->num_trans();
  double unf = 0;
  for (j=1; j<=n; j++)
    if (j != trans)
      if (abs(freq[trans] - freq[j]) <
	  fi->min_dist(fi->cell(trans),fi->cell(j))) 
	unf +=1;
  return unf;
}


/// changes frequency of `trans' to `f' only if this improves
/// objective function.
/// @return true if there is an improvement, false otherwise.
bool FAP_sol::change_freq_if_improves(int trans, int f)
{
  assert(trans > 0 && trans <= fi->num_trans());
  assert(f > 0 && f <= fi->num_freq());
  if (freq[trans] == f)
    return false;
  int old_freq = freq[trans];

  // calculate the interference and unfeasibility due to
  // this transmitter
  double old_int = interference_due(trans);
  double old_unf = unfeasibility_due(trans);
  
  // we try to change frequency of `trans' and see what heapens...
  freq[trans] = f;
  double new_int = interference_due(trans);
  double new_unf = unfeasibility_due(trans);

  // compare the values
  if (of_less(old_unf, old_int, new_unf, new_int)) {
    // change the values for objective function
    interf -= (old_int - new_int);
    unf -= (long int)(old_unf - new_unf);//(cout<<'+').flush();
    return true;
  } else {
    // return to the old value
    freq[trans] = old_freq;//(cout<<'*').flush();
    return false;
  }
}

/// This is a parser for the assignment format for FAP
// instances. It reads a solutions from input stream `is'.
void FAP_sol::input_sol(ifstream &is)
{
  char *to_match, ident[MAX_IDENT_SIZE];
  int cell, channel;
  float dummy_f;

  m(FORMAT) m({)  m(TYPE) m(ASSIGNMENT) m(;)m(VERSION);
  if (!read_float(&dummy_f,is) || dummy_f != 1.0) {
    cerr << "wrong version of format.!\n";
    exit(1);
  }
  m(;) match("}");
  m(GENERAL_INFORMATION)match("{");
  m(SCENARIO_ID);
  readw; m(;);
  m(NAME);
  eat_text(is); m(;);
  if (read_token("ANNOTATION",is)) {
    eat_text(is);
    m(;);
  }
  match("}");
  
  m(CELLS);match("{");
  while (read_integer(&cell, is)) {
    match("{"); match("(");
    readi(channel);match(",");
    if (!(read_token("0", is) || read_token("+", is))) {
      cerr << "error reading assignment format.\n";
      exit(1);
    }
    assert(cell >= 0 && cell < fi->num_trans());
    freq[cell+1] = channel - fi->get_spectrum_ini() + 1;
    match(")");m(;);match("}");
  }
  match("}");
  // compute the objective funcion
  calc_of();
  return;

  error_integer:
  error_ident:
  error_match:
  cerr << "error reading assignment format.\n";
}


  
